import asyncio
import dataclasses
import functools
import logging
from typing import Any, Awaitable, Dict, List, Optional, Set, Tuple, Union

import aiobservable
import yarl
from autobahn import wamp
from autobahn.asyncio.component import Component

from .format import human_repr, indent_multiline

__all__ = ["ConnectionConfig", "SubscriptionEvent", "Connection",
           "get_transports"]

log = logging.getLogger(__name__)


@dataclasses.dataclass()
class ConnectionConfig:
    realm: str
    transports: Union[str, List[dict]]
    subscriptions: Set[str] = dataclasses.field(default_factory=set)

    def __str__(self) -> str:
        return f"(realm={self.realm}, endpoint={self.endpoint})"

    @property
    def endpoint(self) -> str:
        if isinstance(self.transports, str):
            return self.transports

        try:
            return self.transports[0]["url"]
        except (IndexError, KeyError):
            raise ValueError("No transport given") from None


def is_transport_lost(e: Exception) -> bool:
    if isinstance(e, wamp.TransportLost):
        log.info("encountered transport lost, treating as fatal to avoid reconnect. %s", e)
        return True

    return False


@dataclasses.dataclass()
class SubscriptionEvent:
    uri: str
    args: Tuple[Any]
    kwargs: Dict[str, Any]

    def __str__(self) -> str:
        fmt = f"{self.uri}"

        args_fmt = self.format_args()
        if args_fmt:
            fmt += f" ({args_fmt})"

        kwargs_fmt = self.format_kwargs()
        if kwargs_fmt:
            fmt += f" *\n{kwargs_fmt}"

        return fmt

    def format_args(self) -> str:
        return ", ".join(indent_multiline(human_repr(arg)) for arg in self.args)

    def format_kwargs(self) -> str:
        return "\n".join(f"  {key} = {indent_multiline(human_repr(value))}"
                         for key, value in self.kwargs.items())


class Connection(aiobservable.Observable):
    config: ConnectionConfig

    _loop: asyncio.get_event_loop()
    _component: Component
    _join_future: asyncio.Future

    _subscriptions: Dict[str, wamp.types.ISubscription]

    def __init__(self, config: ConnectionConfig, *, loop: asyncio.AbstractEventLoop = None) -> None:
        super().__init__()

        self.config = config
        self._component = Component(
            realm=config.realm,
            transports=config.transports,
            # issue! see https://github.com/crossbario/autobahn-python/issues/1231
            is_fatal=is_transport_lost,
        )
        self.__add_component_listeners()

        self._loop = loop or asyncio.get_event_loop()
        self._join_future = self._loop.create_future()

        self._subscriptions = {}

    def __repr__(self) -> str:
        return f"Connection({self.config!r})"

    def __str__(self) -> str:
        return f"Connection({self.config})"

    @property
    def component(self) -> Component:
        return self._component

    def __add_component_listeners(self) -> None:
        component = self._component

        @component.on_connect
        def on_connect(*_) -> None:
            log.debug("%s: connected", self)

        @component.on_connectfailure
        def on_fail(_, e: Exception) -> None:
            log.debug("%s: failed to connect: %s", self, e)
            try:
                self._join_future.set_exception(e)
            except asyncio.InvalidStateError:
                pass

        @component.on_disconnect
        def on_disconnect(_, *, was_clean: bool) -> None:
            log.debug("%s: disconnected clean=%s", self, was_clean)

        @component.on_join
        def on_join(*_) -> None:
            log.debug("%s joined", self)

            try:
                self._join_future.set_result(None)
            except asyncio.InvalidStateError:
                pass

        @component.on_ready
        async def on_ready(*_) -> None:
            coro_gen = (self.add_subscription(topic) for topic in self.config.subscriptions)
            await asyncio.gather(*coro_gen)

        @component.on_leave
        def on_leave(*_) -> None:
            log.debug("%s: left session", self)
            self._loop.create_future()

    async def __on_event(self, topic: str, *args, **kwargs) -> None:
        await self.emit(SubscriptionEvent(topic, args, kwargs))

    def has_subscription(self, topic: str) -> bool:
        return topic in self.config.subscriptions

    async def _subscribe(self, topic: str) -> None:
        session = await self.session
        self._subscriptions[topic] = await session.subscribe(
            functools.partial(self.__on_event, topic),
            topic,
        )

    async def add_subscription(self, topic: str) -> None:
        if self.has_subscription(topic):
            return

        self.config.subscriptions.add(topic)

        if self.connected:
            await self._subscribe(topic)
        else:
            log.debug("%s: not connected yet, subscribing when ready", self)

    async def remove_subscription(self, topic: str) -> None:
        self.config.subscriptions.discard(topic)

        try:
            subscription = self._subscriptions.pop(topic)
        except KeyError:
            return

        await subscription.unsubscribe()

    @property
    def component_session(self) -> Optional[wamp.ISession]:
        return self._component._session

    @property
    def connected(self) -> bool:
        session = self.component_session
        return session and session.is_connected()

    @property
    def session(self) -> Awaitable[wamp.ISession]:
        return self._loop.create_task(self.get_session())

    async def get_session(self) -> wamp.ISession:
        if not self.connected:
            await self.open()

        session = self.component_session
        assert session, "session should not be None at this point"

        return session

    async def open(self) -> None:
        log.debug("%s: opening connection", self)

        # some exceptions can only be captured by awaiting this,
        # however we don't want to wait until the component is "done"
        # (i.e. closed), so we race it with joining the session.
        done_fut = self._component.start(loop=self._loop)

        done_fs, _ = await asyncio.wait(
            (done_fut, self._join_future),
            return_when=asyncio.FIRST_COMPLETED
        )

        fut = done_fs.pop()

        try:
            return fut.result()
        except Exception:
            await self.close()
            raise

    async def close(self) -> None:
        if not self.connected:
            log.debug("%s: already closed", self)
            return

        log.debug("%s: closing connection", self)
        fut = self.component_session.leave()
        if fut:
            await fut

        await self.wait_done()

    async def wait_done(self) -> None:
        done_f = self._component._done_f
        if done_f:
            try:
                await done_f
            except Exception:
                log.exception("error happened cool")


def get_transports(url: Union[str, yarl.URL]) -> List[dict]:
    url = yarl.URL(url)

    # autobahn python doesn't understand the tcp scheme
    if url.scheme == "tcp":
        url = url.with_scheme("rs")

    if url.scheme in ("rs", "rss"):
        t_type = "rawsocket"
    else:
        t_type = "websocket"

    return [{
        "type": t_type,
        "url": str(url),
    }]



import dataclasses
from typing import Any

import yaml

__all__ = ["human_repr", "human_result", "indent_multiline"]


def human_repr(o: Any) -> str:
    """Convert the given object to a meaningful representation.

    The output is similar to YAML.
    """
    if dataclasses.is_dataclass(o):
        return str(o)

    try:
        return yaml.dump(o)
    except yaml.YAMLError:
        return str(o)


def human_result(result: Any) -> str:
    if result is None:
        return "ok"

    return indent_multiline(human_repr(result))


def indent_multiline(s: str, indentation: str = "  ", add_newlines: bool = True) -> str:
    lines = s.splitlines()
    if len(lines) <= 1:
        return s

    lines_str = "\n".join(f"{indentation}{line}" for line in lines)
    if add_newlines:
        return f"\n{lines_str}\n"
    else:
        return lines_str



import argparse
import asyncio
import signal
import sys
from typing import Any, Awaitable, Callable

from autobahn import wamp

import libwampli
import wampli


def get_parser() -> argparse.ArgumentParser:
    """Create the argument parser for the wampli cli.

    Resulting args contain the special value `entrypoint_func` which
    is either a function which should be called with the args or `None` if no
    subcommand was selected.
    """
    parser = argparse.ArgumentParser("wampli", description="A command line interface for the WAMP protocol.")
    parser.set_defaults(entrypoint_func=None)

    # CURRENTLY UNUSED
    # parser.add_argument("-c", "--config", help="location of the config file")
    # parser.add_argument("--ignore-config", action="store_true", default=False,
    #                     help="don't search for a config file")

    def add_wamp_argument_group(p) -> None:
        wamp = p.add_argument_group("wamp arguments", description="Arguments for WAMP")
        wamp.add_argument("-u", "--url", help="url of the WAMP router to connect to", required=True)
        wamp.add_argument("-r", "--realm", help="realm to join", required=True)

    def add_uri_argument(p, **kwargs) -> None:
        p.add_argument("uri", help="WAMP URI", **kwargs)

    def add_procedure_args_arguments(p) -> None:
        add_uri_argument(p)
        p.add_argument("args",
                       help="Arguments to provide. Use key=value for keyword arguments",
                       nargs="*")

    add_wamp_argument_group(parser)

    subparsers = parser.add_subparsers(title="commands")

    call = subparsers.add_parser("call", help="call a procedure")
    call.set_defaults(entrypoint_func=_call_cmd)
    add_procedure_args_arguments(call)

    publish = subparsers.add_parser("publish", help="publish to a topic")
    publish.set_defaults(entrypoint_func=_publish_cmd)
    add_procedure_args_arguments(publish)

    subscribe = subparsers.add_parser("subscribe", help="subscribe to a topic")
    subscribe.set_defaults(entrypoint_func=_subscribe_cmd)
    add_uri_argument(subscribe, nargs="*")

    shell = subparsers.add_parser("shell", help="start the interactive shell")
    shell.set_defaults(entrypoint_func=_shell_cmd)

    return parser


def get_connection_config(args: argparse.Namespace) -> libwampli.ConnectionConfig:
    transports = libwampli.get_transports(args.url)
    return libwampli.ConnectionConfig(realm=args.realm, transports=transports)


def get_session_context(args: argparse.Namespace) -> libwampli.SessionContext:
    return libwampli.SessionContext(get_connection_config(args))


def _run_async(loop: asyncio.AbstractEventLoop, coro: Awaitable) -> Any:
    def graceful_exit() -> None:
        pending = asyncio.Task.all_tasks()
        for task in pending:
            task.cancel()

        loop.stop()

    try:
        loop.add_signal_handler(signal.SIGINT, graceful_exit)
        loop.add_signal_handler(signal.SIGTERM, graceful_exit)
    except NotImplementedError:
        pass

    try:
        return loop.run_until_complete(coro)
    except KeyboardInterrupt:
        print("Exiting")

    loop.close()


def _run_async_cmd(cmd: Callable[[], Any]) -> Any:
    loop = asyncio.get_event_loop()
    _run_async(loop, cmd())


def _run_cmd(cmd: Callable[[], Any]) -> None:
    try:
        result = _run_async_cmd(cmd)
    except wamp.ApplicationError as e:
        sys.exit(e.error_message())
    else:
        if result is None:
            print("done")
        else:
            print(libwampli.human_result(result))


def _call_cmd(args: argparse.Namespace) -> None:
    async def cmd() -> None:
        async with get_session_context(args) as session:
            return await session.call(args.uri, *call_args, **call_kwargs)

    call_args, call_kwargs = libwampli.parse_args(args.args)
    _run_cmd(cmd)


def _publish_cmd(args: argparse.Namespace) -> None:
    async def cmd() -> None:
        async with get_session_context(args) as session:
            # TODO provide options for acknowledge and so on
            ack = session.publish(args.uri, *publish_args, **publish_kwargs)
            if ack is not None:
                return await ack
            else:
                return None

    publish_args, publish_kwargs = libwampli.parse_args(args.args)
    _run_cmd(cmd)


def _subscribe_cmd(args: argparse.Namespace) -> None:
    async def on_event(event: libwampli.SubscriptionEvent) -> None:
        print(event)

    async def cmd() -> None:
        session_context = get_session_context(args)

        session_context.on(libwampli.SubscriptionEvent, on_event)

        async with session_context:
            coro_gen = (session_context.add_subscription(uri) for uri in args.uri)
            await asyncio.gather(*coro_gen)
            print(f"subscribed to {len(args.uri)} topic(s)")

            await session_context.wait()

    _run_cmd(cmd)


def _shell_cmd(args: argparse.Namespace) -> None:
    shell = wampli.Shell(get_connection_config(args))
    shell.run()


def main() -> None:
    """Entry point for the cli"""
    parser = get_parser()
    args = parser.parse_args()
    func = args.entrypoint_func

    if func is None:
        parser.print_usage()
        return

    func(args)


if __name__ == "__main__":
    main()



import io
import re
import shlex
import tokenize
from typing import Any, Dict, Iterable, List, MutableSequence, Pattern, Tuple, Union

import yaml

__all__ = ["parse_arg_value", "parse_args",
           "split_function_style", "split_arg_string",
           "ready_uri"]

# match: wamp.session.get(12345, key=value)
RE_FUNCTION_STYLE: Pattern = re.compile(
    r"""^
      ((?: (?: [0-9a-z_]+\. ) | \. )* (?: [0-9a-z_]+ )?)    # match URI (1)
      \s?
      \(
        (.*)                                                # arguments (2)
      \)
    $""",
    re.VERBOSE,
)


def split_function_style(text: str) -> List[str]:
    """Split a function style call text representation into its arguments.
    Returns:
        Empty list if the given string didn't match the function style,
        otherwise a list with at least the URI as its first item.
    """
    match = RE_FUNCTION_STYLE.match(text)
    if match is None:
        return []

    uri, arg_string = match.groups()
    args = [uri]

    if arg_string:
        token_gen = tokenize.generate_tokens(io.StringIO(arg_string).readline)
        # get the indices of the commas in the string
        commapos = (
            -1,
            *(token.end[1] for token in token_gen if token.string == ","),
            len(arg_string) + 1,
        )

        args.extend([
            arg_string[commapos[i] + 1: commapos[i + 1] - 1]
            for i in range(len(commapos) - 1)
        ])

    return args


def split_arg_string(arg: str) -> List[str]:
    """Split an argument string into its arguments"""
    res = split_function_style(arg)
    return res or shlex.split(arg)


def parse_arg_value(val: str) -> Any:
    """Parse a string value into its Python representation."""
    return yaml.safe_load(val)


# match: key=value
RE_KWARGS_MATCH: Pattern = re.compile(r"^([a-z][a-z0-9_]{2,})\s*=(.*)$")


def parse_args(args: Union[Iterable[str], str]) -> Tuple[List[Any], Dict[str, Any]]:
    """Parse string arguments into their Python representation.
    Returns:
        2-tuple (args, kwargs) where the first item is a `list` containing
        the positional arguments and the second item a `dict` containing
        the keyword arguments (key=value).
    """
    if isinstance(args, str):
        args = split_arg_string(args)

    _args: List[Any] = []
    _kwargs: Dict[str, Any] = {}

    for arg in args:
        match = RE_KWARGS_MATCH.match(arg)

        if match is None:
            _args.append(parse_arg_value(arg))
        else:
            key, value = match.groups()
            _kwargs[key] = parse_arg_value(value)

    return _args, _kwargs


def ready_uri(args: MutableSequence[Any]) -> None:
    try:
        uri = args[0]
    except IndexError:
        raise IndexError("Please provide a URI")

    if not isinstance(uri, str):
        raise TypeError("URI must be a string")



def main() -> None:
    """Run the wampli.cli.main function.

    In addition, make it possible to run without the -m flag.
    """
    try:
        import wampli.cli
    except ImportError:
        import sys
        from os import path

        sys.path.append(path.abspath(path.join(__file__, "../..")))

        import wampli.cli

    wampli.cli.main()


if __name__ == "__main__":
    main()



import asyncio
import cmd
import dataclasses
import queue
import textwrap
import threading
from typing import Any, Iterable, Mapping, NoReturn, Optional

import txaio
from autobahn.wamp import ApplicationError

import libwampli

__all__ = ["Shell"]

# special signal to indicate that the worker thread should stop
STOP_SIGNAL = object()


@dataclasses.dataclass()
class Task:
    action: str
    args: Iterable[Any] = dataclasses.field(default_factory=list)
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)


def worker(config: libwampli.ConnectionConfig, receive: queue.Queue) -> None:
    """Thread worker which performs async tasks.

    You can send `STOP_SIGNAL` to the `receive` queue to stop the worker.

    Args:
        config: Connection config to create the connection from
        receive: Queue to get new tasks from.
            As soon as the worker is running it will pull `Task` instances
            from the queue and execute them. Each time a task is finished
            `queue.Queue.task_done()` is called.
            The special "task" `STOP_SIGNAL` stops the worker.
    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    txaio.config.loop = loop

    connection = libwampli.Connection(config)

    async def on_subscription_event(event: libwampli.SubscriptionEvent) -> None:
        print(f"received event: {event}")

    connection.on(libwampli.SubscriptionEvent, on_subscription_event)

    async def handle_task(task: Task) -> None:
        session = await connection.session

        if task.action == "call":
            try:
                result = await session.call(*task.args, **task.kwargs)
            except ApplicationError as e:
                print(e.error_message())
            else:
                print(libwampli.human_result(result))
        elif task.action == "publish":
            ack = session.publish(*task.args, **task.kwargs)
            if ack is not None:
                await ack

            print("done")
        elif task.action == "subscribe":
            try:
                topic = next(iter(task.args))
            except StopIteration:
                print("no topic provided")
            else:
                if connection.has_subscription(topic):
                    print(f"already subscribed to {topic}")
                else:
                    await connection.add_subscription(topic)
        elif task.action == "unsubscribe":
            try:
                topic = next(iter(task.args))
            except StopIteration:
                print("no topic provided")
            else:
                if not connection.has_subscription(topic):
                    print(f"not subscribed to {topic}")
                else:
                    await connection.remove_subscription(topic)
        else:
            print(f"unknown task given to worker: {task}")

    async def runner() -> NoReturn:
        print("connecting...")
        await connection.open()
        print("connected")

        async def _handle_task(_task: Task) -> None:
            try:
                await handle_task(_task)
            finally:
                receive.task_done()

        while True:
            task = await loop.run_in_executor(None, receive.get)

            if task is STOP_SIGNAL:
                print("stopping worker")
                break
            else:
                loop.create_task(_handle_task(task))

        print("waiting for connection to close!")
        await connection.close()

        this_task = asyncio.current_task()
        tasks = asyncio.all_tasks(loop)
        for task in tasks:
            if task is not this_task:
                task.cancel()

    loop.run_until_complete(runner())
    loop.close()


# TODO alias uri

class Shell(cmd.Cmd):
    intro = textwrap.dedent("""
        Type 'help' or '?' to list all commands. 
        Use 'exit' to exit the shell.
    """).strip()

    prompt = "(WAMPli) "

    _connection_config: libwampli.ConnectionConfig

    _send_queue: queue.Queue
    _receive_queue: queue.Queue
    _worker_thread: Optional[threading.Thread]

    def __init__(self, config: libwampli.ConnectionConfig) -> None:
        super().__init__()

        self._connection_config = config

        self._send_queue = queue.Queue()
        self._receive_queue = queue.Queue()

    @property
    def worker_running(self) -> bool:
        return self._worker_thread and self._worker_thread.is_alive()

    def run(self) -> None:
        self._start_worker()

        try:
            self.cmdloop()
        finally:
            self._stop_worker()

    def _start_worker(self) -> None:
        thread = threading.Thread(
            name="Shell worker",
            target=worker,
            args=(self._connection_config, self._send_queue),
        )

        self._worker_thread = thread
        self._worker_thread.start()

    def _stop_worker(self) -> None:
        if self.worker_running:
            self._send_queue.put(STOP_SIGNAL)
            self._worker_thread.join(timeout=20)

    def do_exit(self, _) -> bool:
        """Exit the shell."""
        print("Goodbye")
        return True

    def default(self, line: str) -> None:
        args = libwampli.split_function_style(line)
        if args:
            args, kwargs = libwampli.parse_args(args)
            self._call(args, kwargs)
        else:
            super().default(line)

    def _call(self, args, kwargs) -> None:
        libwampli.ready_uri(args)

        task = Task("call", args, kwargs)
        self._send_queue.put_nowait(task)

    def do_call(self, arg: str) -> None:
        """Call a procedure."""
        args, kwargs = libwampli.parse_args(arg)
        self._call(args, kwargs)

    def do_publish(self, arg: str) -> None:
        """Publish to a topic."""
        args, kwargs = libwampli.parse_args(arg)
        libwampli.ready_uri(args)

        task = Task("publish", args, kwargs)
        self._send_queue.put_nowait(task)

    def do_subscribe(self, arg: str) -> None:
        """Subscribe to a topic"""
        args = [arg]
        libwampli.ready_uri(args)

        task = Task("subscribe", args)
        self._send_queue.put_nowait(task)

    def do_unsubscribe(self, arg: str) -> None:
        """Unsubscribe from a topic"""
        args = [arg]
        libwampli.ready_uri(args)

        task = Task("unsubscribe", args)
        self._send_queue.put_nowait(task)



import re

import setuptools

# we can't import the module because we most certainly don't have all the
# dependencies installed. So let's just grab the text content and do some
# regex magic.
with open("libwampli/__init__.py", "r") as f:
    content = f.read()

# we're specifically interested in __magic__ attributes like __version__
# and __author__. This regex matches them and their raw values.
matches = re.findall(r"^(__\w+__) = (.+)$", content, re.MULTILINE)
# eval the raw value to get the Python rep
wampli = {key: eval(value) for key, value in matches}

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="wampli",
    version=wampli["__version__"],
    author=wampli["__author__"],
    author_email="team@giesela.dev",
    url="https://github.com/gieseladev/wampli",

    licence="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",

    packages=setuptools.find_packages(exclude=("docs", "tests")),

    python_requires="~=3.7",
    install_requires=[
        "aiobservable",
        "autobahn",
        "cbor",
        "konfi",
        "pyyaml",
        "yarl",
    ],

    entry_points={
        "console_scripts": [
            "wampli=wampli.cli:main",
        ],
    },
)



import asyncio

from autobahn.asyncio.component import Component
from autobahn.wamp.interfaces import ISession

from .connection import Connection

__all__ = ["SessionContext", "wait_for_leave"]


class SessionContext(Connection):
    async def __aenter__(self) -> ISession:
        await self.open()
        return await self.session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def wait(self) -> None:
        await wait_for_leave(self._component, loop=self._loop)


def wait_for_leave(c: Component, *, loop: asyncio.AbstractEventLoop) -> asyncio.Future:
    fut = loop.create_future()

    @c.on_leave
    def on_leave(*_) -> None:
        fut.set_result(None)

    return fut
